@@ -15,8 +15,6 @@ from greenlet import greenlet as RawGreenlet
 from greenlet import getcurrent
 from greenlet import GreenletExit
 
-
-
 __all__ = [
     'getcurrent',
     'GreenletExit',
@@ -58,6 +56,7 @@ from gevent.exceptions import LoopExit
 
 from gevent._waiter import Waiter
 
+
 # Need the real get_ident. We're imported early enough (by gevent/__init__.py)
 # that we can be sure nothing is monkey patched yet.
 get_thread_ident = __import__(thread_mod_name).get_ident
@@ -223,7 +222,8 @@ class signal(object):
 
     This returns an object with the useful method ``cancel``, which,
     when called, will prevent future deliveries of *signalnum* from
-    calling *handler*.
+    calling *handler*. It's best to keep the returned object alive
+    until you call ``cancel``.
 
     .. note::
 
@@ -236,6 +236,10 @@ class signal(object):
 
         The ``handler`` argument is required to
         be callable at construction time.
+
+    .. versionchanged:: 20.5.1
+       The ``cancel`` method now properly cleans up all native resources,
+       and drops references to all the arguments of this function.
     """
     # This is documented as a function, not a class,
     # so we're free to change implementation details.
@@ -248,27 +252,39 @@ class signal(object):
 
         self.hub = _get_hub_noargs()
         self.watcher = self.hub.loop.signal(signalnum, ref=False)
-        self.watcher.start(self._start)
         self.handler = handler
         self.args = args
         self.kwargs = kwargs
         if self.greenlet_class is None:
             from gevent import Greenlet
+            type(self).greenlet_class = Greenlet
             self.greenlet_class = Greenlet
 
-    def _get_ref(self):
-        return self.watcher.ref
+        self.watcher.start(self._start)
 
-    def _set_ref(self, value):
-        self.watcher.ref = value
-
-    ref = property(_get_ref, _set_ref)
-    del _get_ref, _set_ref
+    ref = property(
+        lambda self: self.watcher.ref,
+        lambda self, nv: setattr(self.watcher, 'ref', nv)
+    )
 
     def cancel(self):
-        self.watcher.stop()
+        if self.watcher is not None:
+            self.watcher.stop()
+            # Must close the watcher at a deterministic time, otherwise
+            # when CFFI reclaims the memory, the native loop might still
+            # have some reference to it; if anything tries to touch it
+            # we can wind up writing to memory that is no longer valid,
+            # leading to a wide variety of crashes.
+            self.watcher.close()
+        self.watcher = None
+        self.handler = None
+        self.args = None
+        self.kwargs = None
+        self.hub = None
+        self.greenlet_class = None
 
     def _start(self):
+        # TODO: Maybe this should just be Greenlet.spawn()?
         try:
             greenlet = self.greenlet_class(self.handle)
             greenlet.switch()
@@ -515,19 +531,23 @@ class Hub(WaitOperationsGreenlet):
         if not issubclass(type, self.NOT_ERROR):
             self.print_exception(context, type, value, tb)
         if context is None or issubclass(type, self.SYSTEM_ERROR):
-            self.handle_system_error(type, value)
+            self.handle_system_error(type, value, tb)
 
-    def handle_system_error(self, type, value):
+    def handle_system_error(self, type, value, tb=None):
         """
         Called from `handle_error` when the exception type is determined
         to be a :attr:`system error <SYSTEM_ERROR>`.
 
         System errors cause the exception to be raised in the main
         greenlet (the parent of this hub).
+
+        .. versionchanged:: 20.5.1
+           Allow passing the traceback to associate with the
+           exception if it is rethrown into the main greenlet.
         """
         current = getcurrent()
         if current is self or current is self.parent or self.loop is None:
-            self.parent.throw(type, value)
+            self.parent.throw(type, value, tb)
         else:
             # in case system error was handled and life goes on
             # switch back to this greenlet as well
@@ -537,7 +557,7 @@ class Hub(WaitOperationsGreenlet):
             except: # pylint:disable=bare-except
                 traceback.print_exc(file=self.exception_stream)
             try:
-                self.parent.throw(type, value)
+                self.parent.throw(type, value, tb)
             finally:
                 if cb is not None:
                     cb.stop()
@@ -618,14 +638,44 @@ class Hub(WaitOperationsGreenlet):
                 loop.run()
             finally:
                 loop.error_handler = None  # break the refcount cycle
+
+            # This function must never return, as it will cause
+            # switch() in the parent greenlet to return an unexpected
+            # value. This can show up as unexpected failures e.g.,
+            # from Waiters raising AssertionError or MulitpleWaiter
+            # raising invalid IndexError.
+            #
+            # It is still possible to kill this greenlet with throw.
+            # However, in that case switching to it is no longer safe,
+            # as switch will return immediately.
+            #
+            # Note that there's a problem with simply doing
+            # ``self.parent.throw()`` and never actually exiting this
+            # greenlet: The greenlet tends to stay alive. This is
+            # because throwing the exception captures stack frames
+            # (regardless of what we do with the argument) and those
+            # get saved. In addition to this object having
+            # ``gr_frame`` pointing to this method, which contains
+            # ``self``, which points to the parent, and both of which point to
+            # an internal thread state dict that points back to the current greenlet for the thread,
+            # which is likely to be the parent: a cycle.
+            #
+            # We can't have ``join()`` tell us to finish, because we
+            # need to be able to resume after this throw. The only way
+            # to dispose of the greenlet is to use ``self.destroy()``.
+
             debug = []
             if hasattr(loop, 'debug'):
                 debug = loop.debug()
-            self.parent.throw(LoopExit('This operation would block forever', self, debug))
-        # this function must never return, as it will cause switch() in the parent greenlet
-        # to return an unexpected value
-        # It is still possible to kill this greenlet with throw. However, in that case
-        # switching to it is no longer safe, as switch will return immediately
+            loop = None
+
+            self.parent.throw(LoopExit('This operation would block forever',
+                                       self,
+                                       debug))
+            # Execution could resume here if another blocking API call is made
+            # in the same thread and the hub hasn't been destroyed, so clean
+            # up anything left.
+            debug = None
 
     def start_periodic_monitoring_thread(self):
         if self.periodic_monitoring_thread is None and GEVENT_CONFIG.monitor_thread:
@@ -648,13 +698,20 @@ class Hub(WaitOperationsGreenlet):
         return self.periodic_monitoring_thread
 
     def join(self, timeout=None):
-        """Wait for the event loop to finish. Exits only when there are
-        no more spawned greenlets, started servers, active timeouts or watchers.
+        """
+        Wait for the event loop to finish. Exits only when there
+        are no more spawned greenlets, started servers, active
+        timeouts or watchers.
 
-        If *timeout* is provided, wait no longer for the specified number of seconds.
+        .. caution:: This doesn't clean up all resources associated
+           with the hub. For that, see :meth:`destroy`.
 
-        Returns True if exited because the loop finished execution.
-        Returns False if exited because of timeout expired.
+        :param float timeout: If *timeout* is provided, wait no longer
+            than the specified number of seconds.
+
+        :return: `True` if this method returns because the loop
+                 finished execution. Or `False` if the timeout
+                 expired.
         """
         assert getcurrent() is self.parent, "only possible from the MAIN greenlet"
         if self.dead:
@@ -668,21 +725,46 @@ class Hub(WaitOperationsGreenlet):
 
         try:
             try:
+                # Switch to the hub greenlet and let it continue.
+                # Since we're the parent greenlet of the hub, when it exits
+                # by `parent.throw(LoopExit)`, control will resume here.
+                # If the timer elapses, however, ``waiter.switch()`` is called and
+                # again control resumes here, but without an exception.
                 waiter.get()
             except LoopExit:
+                # Control will immediately be returned to this greenlet.
                 return True
         finally:
+            # Clean up as much junk as we can. There is a small cycle in the frames,
+            # and it won't be GC'd.
+            # this greenlet -> this frame
+            # this greenlet -> the exception that was thrown
+            # the exception that was thrown -> a bunch of other frames, including this frame.
+            # some frame calling self.run() -> self
+            del waiter # this frame -> waiter -> self
+            del self # this frame -> self
             if timeout is not None:
                 timeout.stop()
                 timeout.close()
+            del timeout
         return False
 
     def destroy(self, destroy_loop=None):
         """
         Destroy this hub and clean up its resources.
 
-        If you manually create hubs, you *should* call this
+        If you manually create hubs, or you use a hub or the gevent
+        blocking API from multiple native threads, you *should* call this
         method before disposing of the hub object reference.
+
+        Once this is done, it is impossible to continue running the
+        hub. Attempts to use the blocking gevent API with pre-existing
+        objects from this native thread and bound to this hub will fail.
+
+        .. versionchanged:: 20.5.1
+            Ensure that Python stack frames and greenlets referenced by this
+            hub are cleaned up. This guarantees that switching to the hub again
+            is not safe after this. (It was never safe, but it's even less safe.)
         """
         if self.periodic_monitoring_thread is not None:
             self.periodic_monitoring_thread.kill()
@@ -693,6 +775,19 @@ class Hub(WaitOperationsGreenlet):
         if self._threadpool is not None:
             self._threadpool.kill()
             del self._threadpool
+
+        # Let the frame be cleaned up by causing the run() function to
+        # exit. This is the only way to guarantee that the hub itself
+        # and the main greenlet, if this was a secondary thread, get
+        # cleaned up. Otherwise there are likely to be reference
+        # cycles still around. We MUST do this before we destroy the
+        # loop; if we destroy the loop and then switch into the hub,
+        # things will go VERY, VERY wrong.
+        try:
+            self.throw(GreenletExit)
+        except LoopExit:
+            pass
+
         if destroy_loop is None:
             destroy_loop = not self.loop.default
         if destroy_loop:
@@ -705,12 +800,12 @@ class Hub(WaitOperationsGreenlet):
             # thread.
             set_loop(self.loop)
 
-
         self.loop = None
         if _get_hub() is self:
             set_hub(None)
 
 
+
     # XXX: We can probably simplify the resolver and threadpool properties.
 
     @property
